/* 
 * tkRoot.c --
 *
 *	This module implements "rootwin" widgets for the Tk
 *	toolkit.  Rootwins are root windows for a given display
 *	and screen. They have a background color, but not much
 *	else in the way of attributes.
 *
 */

/*
#include "default.h"
#include "tkPort.h"
*/
#include "tk.h"
#include "rootwinInt.h"

/*
 * A data structure of the following type is kept for each
 * rootwin that currently exists for this process:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the rootwin.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up. */
    Display *display;		/* Display containing widget.  Used, among
				 * other things, so that resources can be
				 * freed even after tkwin has gone away. */
    Tcl_Interp *interp;		/* Interpreter associated with widget.  Used
				 * to delete widget command. */
    Tcl_Command widgetCmd;	/* Token for rootwin's widget command. */
    char *className;		/* Class name for widget (from configuration
				 * option).  Malloc-ed. */
    char *screenName;		/* Screen on which widget is created.
				 * Malloc-ed, may be NULL. */
    char *visualName;		/* Textual description of visual for window,
				 * from -visual option.  Malloc-ed, may be
				 * NULL. */
    char *colormapName;		/* Textual description of colormap for window,
				 * from -colormap option.  Malloc-ed, may be
				 * NULL. */
    Colormap colormap;		/* If not None, identifies a colormap
				 * allocated for this window, which must be
				 * freed when the window is deleted. */
    XColor *bgColorPtr;		/* If not NULL, this is the window's
				 * background color. */
    Pixmap bgPixmap;		/* If not "None", this is the window's
				 * background pixmap. */
    Tk_Image bgImage;		/* If not NULL, this is the window's
				 * background image. (This will be
				 * copied into "bgPixmap".) */
    char *bgImageName;		/* If not NULL, this is the name of
				 * the window's background image. */
    Tk_Cursor cursor;		/* Current cursor for window, or None. */
    char *takeFocus;		/* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL. */
    int flags;			/* Various flags;  see below for
				 * definitions. */
} Rootwin;

/*
 * Flag bits for rootwins:
 *
 * REDRAW_PENDING:		Non-zero means a DoWhenIdle handler
 *				has already been queued to redraw
 *				this window.
 * GOT_FOCUS:			Non-zero means this widget currently
 *				has the input focus.
 */

#define REDRAW_PENDING		1
#define GOT_FOCUS		4

/*
 * The following flag bits are used so that there can be separate
 * defaults for some configuration options for rootwins and toplevels.
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_COLOR, "-background", "background", "Background",
	DEF_ROOTWIN_BG_COLOR, Tk_Offset(Rootwin, bgColorPtr),
	TK_CONFIG_COLOR_ONLY|TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-background", "background", "Background",
	DEF_ROOTWIN_BG_MONO, Tk_Offset(Rootwin, bgColorPtr),
	TK_CONFIG_MONO_ONLY|TK_CONFIG_NULL_OK},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_STRING, "-class", "class", "Class",
	DEF_ROOTWIN_CLASS, Tk_Offset(Rootwin, className), 0},
    {TK_CONFIG_STRING, "-colormap", "colormap", "Colormap",
	DEF_ROOTWIN_COLORMAP, Tk_Offset(Rootwin, colormapName),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	DEF_ROOTWIN_CURSOR, Tk_Offset(Rootwin, cursor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-screen", "screen", "Screen",
	DEF_ROOTWIN_SCREEN, Tk_Offset(Rootwin, screenName),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-takefocus", "takeFocus", "TakeFocus",
	DEF_ROOTWIN_TAKE_FOCUS, Tk_Offset(Rootwin, takeFocus),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-visual", "visual", "Visual",
	DEF_ROOTWIN_VISUAL, Tk_Offset(Rootwin, visualName),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-tile", "tile", "Tile",
	NULL, Tk_Offset(Rootwin, bgImageName),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};


/*
 *--------------------------------------------------------------
 *
 * Tkwm_XErrorHandler --
 *
 *	Tkwm_XErrorHandler is an error handler, to be installed
 *	via Tk_CreateErrorHandler, that will set a flag if an
 *	X error occurred.
 *
 * Results:
 *	Always returns 0, indicating that the X error was
 *	handled.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
Tkwm_XErrorHandler (clientData, errEventPtr)
    ClientData clientData;	/* Pointer to flag we set	*/
    XErrorEvent *errEventPtr;	/* X error info			*/
{
    int *error;

    error = (int *) clientData;
    *error = 1;
    return 0;
}

/*
 * Forward declarations for procedures defined later in this file:
 */

static int		ConfigureRootwin _ANSI_ARGS_((Tcl_Interp *interp,
			    Rootwin *rootwinPtr, int argc, char **argv,
			    int flags));
static void		DestroyRootwin _ANSI_ARGS_((char *memPtr));
static void		DisplayRootwin _ANSI_ARGS_((ClientData clientData));
static void		RootwinCmdDeletedProc _ANSI_ARGS_((
			    ClientData clientData));
static void		RootwinEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		RootwinWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
static void		MapRootwin _ANSI_ARGS_((ClientData clientData));
static Window		RootCreateProc _ANSI_ARGS_((Tk_Window tkwin,
			    Window parent, ClientData instanceData));
static int		Tkwm_GetDrawableGeometry _ANSI_ARGS_((Display *display,
			    Drawable drawable, Tkwm_DrawableGeometry *geomPtr));

/*
 *--------------------------------------------------------------
 *
 * RootwinCmd --
 *
 *	This procedure is invoked to process the "rootwin"
 *	Tcl command.  See the user documentation for
 *	details on what they do.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static Tk_ClassProcs classProcs = {
    sizeof(Tk_ClassProcs),
    NULL,
    RootCreateProc,
    NULL
};

static int
RootwinCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter.
				 * If we're called by Tk_Init to create a
				 * new application, then this is NULL. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin;
    Rootwin *rootwinPtr;
    Tk_Window new = NULL;
    char *className, *screenName, *visualName, *colormapName, *arg;
    int i, c, length, depth;
    Colormap colormap;
    Visual *visual;
    Window rootWin, parentWin, *childList;
    int nChildren, status;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_MainWindow(interp);

    /*
     * Pre-process the argument list.  Scan through it to find any
     * "-class", "-screen", "-visual", and "-colormap" options.  These
     * arguments need to be processed specially, before the window
     * is configured using the usual Tk mechanisms.
     */

    className = colormapName = screenName = visualName = NULL;
    colormap = None;
    for (i = 2; i < argc; i += 2) {
	arg = argv[i];
	length = strlen(arg);
	if (length < 2) {
	    continue;
	}
	c = arg[1];
	if ((c == 'c') && (strncmp(arg, "-class", strlen(arg)) == 0)
		&& (length >= 3)) {
	    className = argv[i+1];
	} else if ((c == 'c')
		&& (strncmp(arg, "-colormap", strlen(arg)) == 0)) {
	    colormapName = argv[i+1];
	} else if ((c == 's')
		&& (strncmp(arg, "-screen", strlen(arg)) == 0)) {
	    screenName = argv[i+1];
	} else if ((c == 'v')
		&& (strncmp(arg, "-visual", strlen(arg)) == 0)) {
	    visualName = argv[i+1];
	}
    }

    /*
     * Create the window, and deal with the special options -classname,
     * -colormap, -screenname, and -visual.  The order here is tricky,
     * because we want to allow values for these options to come from
     * the database, yet we can't do that until the window is created.
     */

    if (screenName == NULL) {
	screenName = "";
    }

    new = Tk_CreateWindowFromPath(interp, tkwin, argv[1], screenName);
    if (new == NULL) {
	goto error;
    }
    if (className == NULL) {
	className = Tk_GetOption(new, "class", "Class");
	if (className == NULL) {
	    className = "Rootwin";
	}
    }
    Tk_SetClass(new, className);
    if (visualName == NULL) {
	visualName = Tk_GetOption(new, "visual", "Visual");
    }
    if (colormapName == NULL) {
	colormapName = Tk_GetOption(new, "colormap", "Colormap");
    }
    if (visualName != NULL) {
	visual = Tk_GetVisual(interp, new, visualName, &depth,
		(colormapName == NULL) ? &colormap : (Colormap *) NULL);
	if (visual == NULL) {
	    goto error;
	}
	Tk_SetWindowVisual(new, visual, depth, colormap);
    }
    if (colormapName != NULL) {
	colormap = Tk_GetColormap(interp, new, colormapName);
	if (colormap == None) {
	    goto error;
	}
	Tk_SetWindowColormap(new, colormap);
    }

    /*
     * Determine the root window ID, and set the window creation
     * proc so that it will return that X window ID.
     */


    status = XQueryTree(Tk_Display(new), Tk_WindowId(tkwin), &rootWin,
		&parentWin, &childList, &nChildren);
    if(status == 0) {
	Tcl_SetResult(interp, "Couldn't find root window", TCL_STATIC);
	Tk_DestroyWindow(new);
	return TCL_ERROR;
    }

    if (nChildren != 0) {
	XFree(childList);
    }

    Tk_SetClassProcs(new, &classProcs, (ClientData) rootWin);

    /*
     * Create the widget record, process configuration options, and
     * create event handlers.  Then fill in a few additional fields
     * in the widget record from the special options.
     */

    rootwinPtr = (Rootwin *) ckalloc(sizeof(Rootwin));
    rootwinPtr->tkwin = new;
    rootwinPtr->display = Tk_Display(new);
    rootwinPtr->interp = interp;
    rootwinPtr->widgetCmd = Tcl_CreateCommand(interp,
	    Tk_PathName(new), RootwinWidgetCmd,
	    (ClientData) rootwinPtr, RootwinCmdDeletedProc);
    rootwinPtr->className = NULL;
    rootwinPtr->screenName = NULL;
    rootwinPtr->visualName = NULL;
    rootwinPtr->colormapName = NULL;
    rootwinPtr->colormap = colormap;
    rootwinPtr->bgColorPtr = NULL;
    rootwinPtr->bgPixmap = None;
    rootwinPtr->bgImage = NULL;
    rootwinPtr->bgImageName = NULL;
    rootwinPtr->cursor = None;
    rootwinPtr->takeFocus = NULL;
    rootwinPtr->flags = 0;
    Tk_CreateEventHandler(new, StructureNotifyMask|KeyPressMask|ExposureMask|SubstructureNotifyMask|FocusChangeMask,
	    RootwinEventProc, (ClientData) rootwinPtr);

    Tk_MakeWindowExist(new);
    if (ConfigureRootwin(interp, rootwinPtr, argc-2, argv+2, 0) != TCL_OK) {
	goto error;
    }
    interp->result = Tk_PathName(new);
    return TCL_OK;

    error:
    if (new != NULL) {
	Tk_DestroyWindow(new);
    }
    return TCL_ERROR;
}

static Window
RootCreateProc(tkwin, parent, instanceData)
    Tk_Window tkwin;
    Window parent;
    ClientData instanceData;
{
    ((Tk_FakeWin *) tkwin)->flags = TK_TOP_HIERARCHY;
    return ((Window) instanceData);
}


/*
 *--------------------------------------------------------------
 *
 * RootwinWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a rootwin widget.  See the user
 *	documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
RootwinWidgetCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Information about rootwin widget. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    register Rootwin *rootwinPtr = (Rootwin *) clientData;
    int result = TCL_OK;
    size_t length;
    int c, i;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData) rootwinPtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if (c == 'c') {
	if ((strncmp(argv[1], "cget", length) == 0) && (length >= 2)) {
	    if (argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " cget option\"",
			(char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	    result = Tk_ConfigureValue(interp, rootwinPtr->tkwin, configSpecs,
		    (char *) rootwinPtr, argv[2], 0);
	} else if ((strncmp(argv[1], "configure", length) == 0)
		&& (length >= 2)) {
	    if (argc == 2) {
		result = Tk_ConfigureInfo(interp, rootwinPtr->tkwin, configSpecs,
		    (char *) rootwinPtr, (char *) NULL, 0);
	    } else if (argc == 3) {
	        result = Tk_ConfigureInfo(interp, rootwinPtr->tkwin, configSpecs,
		    (char *) rootwinPtr, argv[2], 0);
	    } else {
	        /*
	         * Don't allow the options -class, -newcmap, -screen,
	         * or -visual to be changed.
	         */

	        for (i = 2; i < argc; i++) {
		    length = strlen(argv[i]);
		    if (length < 2) {
		        continue;
		    }
		    c = argv[i][1];
		    if (((c == 'c') && (strncmp(argv[i], "-class", length) == 0)
			    && (length >= 2))
			    || ((c == 'c')
			    && (strncmp(argv[i], "-colormap", length) == 0))
			    || ((c == 's')
			    && (strncmp(argv[i], "-screen", length) == 0))
			    || ((c == 'v')
			    && (strncmp(argv[i], "-visual", length) == 0))) {
		        Tcl_AppendResult(interp, "can't modify ", argv[i],
			        " option after widget is created", (char *) NULL);
		        result = TCL_ERROR;
		        goto done;
		    }
	        }
	        result = ConfigureRootwin(interp, rootwinPtr, argc-2, argv+2,
		        TK_CONFIG_ARGV_ONLY);
	    }
	} else {
	    Tcl_AppendResult(interp, "bad option \"", argv[1],
		    "\"; see the manual page for valid options", (char *) NULL);
	    result = TCL_ERROR;
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\"; see the manual page for valid options", (char *) NULL);
	result = TCL_ERROR;
    }

    done:
    Tcl_Release((ClientData) rootwinPtr);
    return result;
}

/*
 *--------------------------------------------------------------
 *
 * Tkwm_GetDrawableGeometry --
 *	Get drawable gemoetry, if drawable exists. This is a
 *	wrapper for XGetGeometry. (We need to install our own
 *	dummy error handler.)
 *
 * Returns:
 *	Zero on failure (e.g. drawable doesn't exist),
 *	non-zero on success
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
Tkwm_GetDrawableGeometry(display,drawable,geomPtr)
    Display *display;
    Drawable drawable;
    Tkwm_DrawableGeometry *geomPtr;
{
    Tk_ErrorHandler handler;
    int gotXError;
    Status status;

    gotXError = 0;
    handler = Tk_CreateErrorHandler(display, -1, -1, -1,
		Tkwm_XErrorHandler, (ClientData) (&gotXError));
    status = XGetGeometry(display, drawable, &(geomPtr->rootwin),
		&(geomPtr->x),&(geomPtr->y),
		&(geomPtr->width),&(geomPtr->height),
		&(geomPtr->border_width),&(geomPtr->depth));
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);
    return status;
}


/*
 *----------------------------------------------------------------------
 *
 * DestroyRootwin --
 *
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a rootwin at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the rootwin is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyRootwin(memPtr)
    char *memPtr;		/* Info about rootwin widget. */
{
    register Rootwin *rootwinPtr = (Rootwin *) memPtr;

    /* Clean up the internal background image stuff. */

    if (rootwinPtr->bgImage != NULL) {
	Tk_FreeImage(rootwinPtr->bgImage);
	rootwinPtr->bgImage = NULL;
    }

    if (rootwinPtr->bgPixmap != None) {
	Tk_FreePixmap(rootwinPtr->display, rootwinPtr->bgPixmap);
	rootwinPtr->bgPixmap = None;
    }

    /* Free up everything else. */

    Tk_FreeOptions(configSpecs, (char *) rootwinPtr, rootwinPtr->display, 0);
    if (rootwinPtr->colormap != None) {
	Tk_FreeColormap(rootwinPtr->display, rootwinPtr->colormap);
    }
    ckfree((char *) rootwinPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureRootwin --
 *
 *	This procedure is called to process an argv/argc list, plus
 *	the Tk option database, in order to configure (or
 *	reconfigure) a rootwin widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as text string, colors, font,
 *	etc. get set for rootwinPtr;  old resources get freed, if there
 *	were any.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureRootwin(interp, rootwinPtr, argc, argv, flags)
    Tcl_Interp *interp;		/* Used for error reporting. */
    register Rootwin *rootwinPtr;	/* Information about widget;  may or may
				 * not already have values for some fields. */
    int argc;			/* Number of valid entries in argv. */
    char **argv;		/* Arguments. */
    int flags;			/* Flags to pass to Tk_ConfigureWidget. */
{
    if (Tk_ConfigureWidget(interp, rootwinPtr->tkwin, configSpecs,
	    argc, argv, (char *) rootwinPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Don't need to do anything here; it's all done in DisplayRootwin. */

    if (!(rootwinPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayRootwin, (ClientData) rootwinPtr);
    }
    rootwinPtr->flags |= REDRAW_PENDING;

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * RootwinBgImageChanged --
 *
 *	This procedure is invoked when the rootwin's background
 *	image changes. It redraws the background pixmap and
 *	and clears the window, so that the background image
 *	shows up.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The background of the window changes.
 *
 *----------------------------------------------------------------------
 */

static void
RootwinBgImageChanged(clientData, x, y, width, height, imageWidth, imageHeight)
    ClientData clientData;	/* rootwin Pointer		*/
    int x;			/* X coord. of upper left corner
				 * of changed part of image	*/
    int y;			/* Y coord. of upper left corner
				 * of changed part of image	*/
    int width;			/* Width of changed part	*/
    int height;			/* Height of changed part	*/
    int imageWidth;		/* (Possibly changed) width of
				 * the image			*/
    int imageHeight;		/* (Possibly changed) height of
				 * the image			*/
{
    register Rootwin *rootwinPtr = (Rootwin *) clientData;
    register Tk_Window tkwin = rootwinPtr->tkwin;
    Tkwm_DrawableGeometry geomInfo;

    /* Resize the background pixmap if the image size changed. */

    if (rootwinPtr->bgPixmap == None) {
	rootwinPtr->bgPixmap = Tk_GetPixmap(Tk_Display(tkwin),
			Tk_WindowId(tkwin), imageWidth, imageHeight,
			Tk_Depth(tkwin));
    } else {
	if(Tkwm_GetDrawableGeometry(Tk_Display(tkwin),
			rootwinPtr->bgPixmap, &geomInfo) == 0) {
	    return;
	}

	if (geomInfo.width != imageWidth || geomInfo.height != imageHeight) {
	    Tk_FreePixmap(Tk_Display(tkwin),  rootwinPtr->bgPixmap);
	    rootwinPtr->bgPixmap = Tk_GetPixmap(Tk_Display(tkwin),
			Tk_WindowId(tkwin), imageWidth, imageHeight,
			Tk_Depth(tkwin));
	}
    }

    /* Copy the changed image into the pixmap, and clear the window. */

    Tk_RedrawImage(rootwinPtr->bgImage, 0, 0,
		imageWidth, imageHeight, rootwinPtr->bgPixmap, 0, 0);
    /* We shouldn't ALWAYS need to do this... */
    XSetWindowBackgroundPixmap(Tk_Display(tkwin),
			Tk_WindowId(tkwin), rootwinPtr->bgPixmap);

    XClearWindow(rootwinPtr->display, Tk_WindowId(rootwinPtr->tkwin));
}


/*
 *----------------------------------------------------------------------
 *
 * DisplayRootwin --
 *
 *	This procedure is invoked to display a rootwin widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the rootwin in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

static void
DisplayRootwin(clientData)
    ClientData clientData;	/* Information about widget. */
{
    register Rootwin *rootwinPtr = (Rootwin *) clientData;
    register Tk_Window tkwin = rootwinPtr->tkwin;
    int w,h;

    rootwinPtr->flags &= ~REDRAW_PENDING;
    if (rootwinPtr->tkwin == NULL) {
	return;
    }

    /*
     * Change the background image or color. We must assume that the
     * background image changed, so we must free up all the resources
     * used previously, and regenerate them from the image name.
     */

    if (rootwinPtr->bgImage != NULL) {
	Tk_FreeImage(rootwinPtr->bgImage);
	rootwinPtr->bgImage = NULL;
    }
    if (rootwinPtr->bgPixmap != None) {
	Tk_FreePixmap(Tk_Display(tkwin), rootwinPtr->bgPixmap);
	rootwinPtr->bgPixmap = None;
    }

    XSetWindowBackgroundPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin), None);

    /* Set the background image or pixel. */

    if (rootwinPtr->bgImageName != NULL) {
	/* Convert image name to Tk_Image */

	rootwinPtr->bgImage = Tk_GetImage(rootwinPtr->interp, tkwin,
			rootwinPtr->bgImageName, RootwinBgImageChanged,
			rootwinPtr);
	if ( rootwinPtr->bgImage == NULL) {
	    return;
	}

	/* Draw the image on the background pixmap. */

	Tk_SizeOfImage(rootwinPtr->bgImage, &w, &h);
	rootwinPtr->bgPixmap = Tk_GetPixmap(Tk_Display(tkwin),
		Tk_WindowId(tkwin), w, h, Tk_Depth(tkwin));
	if (rootwinPtr->bgPixmap == None) {
	    return;
	}

	Tk_RedrawImage(rootwinPtr->bgImage, 0, 0,
		w, h, rootwinPtr->bgPixmap, 0, 0);

	/* Make sure that the window uses this background pixmap. */

	XSetWindowBackgroundPixmap(Tk_Display(tkwin),
		Tk_WindowId(tkwin), rootwinPtr->bgPixmap);
    } else if (rootwinPtr->bgColorPtr != NULL) {
	Tk_SetWindowBackground(tkwin, rootwinPtr->bgColorPtr->pixel);
    }

    /* Make sure we use the current background. */

    XClearWindow(rootwinPtr->display, Tk_WindowId(rootwinPtr->tkwin));
}

/*
 *--------------------------------------------------------------
 *
 * RootwinEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher on
 *	structure changes to a rootwin.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
RootwinEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    register XEvent *eventPtr;	/* Information about event. */
{
    register Rootwin *rootwinPtr = (Rootwin *) clientData;

    if (((eventPtr->type == Expose) && (eventPtr->xexpose.count == 0))
	    || (eventPtr->type == ConfigureNotify)) {
	goto redraw;
    } else if (eventPtr->type == DestroyNotify) {
	XDestroyWindowEvent *destroyPtr;
	destroyPtr = (XDestroyWindowEvent *) eventPtr;
	if (rootwinPtr->tkwin != NULL
		&& destroyPtr->window == Tk_WindowId(rootwinPtr->tkwin)) {
	    rootwinPtr->tkwin = NULL;
	    Tcl_DeleteCommandFromToken(rootwinPtr->interp, rootwinPtr->widgetCmd);
	    if (rootwinPtr->flags & REDRAW_PENDING) {
		Tcl_CancelIdleCall(DisplayRootwin, (ClientData) rootwinPtr);
	    }
	    Tcl_CancelIdleCall(MapRootwin, (ClientData) rootwinPtr);
	    Tcl_EventuallyFree((ClientData) rootwinPtr, DestroyRootwin);
	}
    } else if (eventPtr->type == FocusIn) {
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    rootwinPtr->flags |= GOT_FOCUS;
	}
    } else if (eventPtr->type == FocusOut) {
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    rootwinPtr->flags &= ~GOT_FOCUS;
	}
    }
    return;

    redraw:
    if ((rootwinPtr->tkwin != NULL) && !(rootwinPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayRootwin, (ClientData) rootwinPtr);
	rootwinPtr->flags |= REDRAW_PENDING;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * RootwinCmdDeletedProc --
 *
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void
RootwinCmdDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    Rootwin *rootwinPtr = (Rootwin *) clientData;
    Tk_Window tkwin = rootwinPtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
	rootwinPtr->tkwin = NULL;
	Tk_DestroyWindow(tkwin);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * MapRootwin --
 *
 *	This procedure is invoked as a when-idle handler to map a
 *	newly-created top-level rootwin.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The rootwin given by the clientData argument is mapped.
 *
 *----------------------------------------------------------------------
 */

static void
MapRootwin(clientData)
    ClientData clientData;		/* Pointer to rootwin structure. */
{
    Rootwin *rootwinPtr = (Rootwin *) clientData;

    /*
     * Wait for all other background events to be processed before
     * mapping window.  This ensures that the window's correct geometry
     * will have been determined before it is first mapped, so that the
     * window manager doesn't get a false idea of its desired geometry.
     */

    Tcl_Preserve((ClientData) rootwinPtr);
    while (1) {
	if (Tcl_DoOneEvent(TCL_IDLE_EVENTS) == 0) {
	    break;
	}

	/*
	 * After each event, make sure that the window still exists
	 * and quit if the window has been destroyed.
	 */

	if (rootwinPtr->tkwin == NULL) {
	    Tcl_Release((ClientData) rootwinPtr);
	    return;
	}
    }
    Tk_MapWindow(rootwinPtr->tkwin);
    Tcl_Release((ClientData) rootwinPtr);
}

int
Rootwin_Init(interp)
    Tcl_Interp *interp;
{
    if(Tcl_PkgRequire(interp, "Tk", "8.4", 0) == NULL) {
        return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, "rootwin", RootwinCmd, NULL, NULL);
    Tcl_PkgProvide(interp, "Rootwin", VERSION);

    return TCL_OK;
}

